package de.mmadler.filecopy.tasks;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.nio.channels.FileChannel;
import java.nio.file.Files;
import java.nio.file.LinkOption;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.mmadler.filecopy.gui.FileCopy;
import javafx.application.Platform;
import javafx.beans.property.SimpleStringProperty;
import javafx.beans.property.StringProperty;
import javafx.concurrent.Task;

public class CopyFile extends Task<Void> {
	private boolean _paused;
	
	private File _file;
	private File _destination;
	
	private long _startTime;
	private Long _chunkSize;
	private StringProperty _percentComplete = new SimpleStringProperty("");
	private StringProperty _timeRunning = new SimpleStringProperty("");
	private StringProperty _copyingSpeed = new SimpleStringProperty("");
	
	private final static double _MAXCHUNKS = 250.0;

	private static Logger logger = null;

	public CopyFile(File file, File destination) {
		_file = file;
		if(destination.isDirectory())
			_destination = new File(destination.getAbsolutePath()+File.separator+_file.getName());
		else
			_destination = destination;

		if(FileCopy.getConfig().writeLog())
			logger = LoggerFactory.getLogger(CopyFile.class);
	}
	
	@Override
	protected Void call() throws Exception {
		Exception copyingException = null;
		
		FileInputStream is = new FileInputStream(_file);
		FileOutputStream os = new FileOutputStream(_destination);
		FileChannel inputChannel = null;
		FileChannel outputChannel = null;
		try {
			inputChannel = is.getChannel();
			outputChannel = os.getChannel();
			_chunkSize = (long) Math.ceil(inputChannel.size() / _MAXCHUNKS);
			int chunks = (int) Math.ceil(inputChannel.size() / _chunkSize.doubleValue());

			updateProgress(0.0, chunks, false);
			for(int chunk=0; chunk<chunks; chunk++) {
				inputChannel.transferTo(chunk*_chunkSize, _chunkSize, outputChannel);
				updateProgress(chunk+1, chunks, false);
				while(isPaused() && !isCancelled())
					Thread.sleep(200);
				if(isCancelled())
					break;
			}
			if (!isCancelled()) {
				succeeded();
				updateProgress(0, chunks, true);
			}
		} catch (Exception e) {
			if(logger!=null) logger.error("Copying of file '"+_file.getAbsolutePath()+"' failed.", e);
			copyingException = e;
			failed();
		} finally {
			inputChannel.close();
			is.close();
			outputChannel.close();
			os.close();
			if(isCancelled() || copyingException!=null) {
				_destination.delete();
				throw copyingException;
			}
			else if(FileCopy.getConfig().keepTimestamp())
				Files.setLastModifiedTime(_destination.toPath(), Files.getLastModifiedTime(_file.toPath(), LinkOption.NOFOLLOW_LINKS));
		}
		
		return null;
	}

	public boolean destinationExists() {
		return _destination.exists();
	}
	
	public boolean destinationWritable() {
		return ((!_destination.exists() && _destination.getParentFile().canWrite()) || _destination.canWrite());
	}
	
	public File renameDestination() {
		int idx = 1;
		do {
			_destination = new File(_destination.getAbsolutePath().replaceFirst("( \\([0-9]+\\))?(\\.[^.]+)?$", " ("+idx+")$2"));
			idx++;
		} while(_destination.exists());
		
		return _destination;
	}
	
	public void setDestination(File destination) {
		_destination = destination.getAbsoluteFile();
	}

	public boolean isPaused() {
		return _paused;
	}

	public void resume() {
		_paused = false;
	}

	public void pause() {
		_paused = true;
	}
	
	private void updateProgress(final double current, double max, boolean reset) {
		if(current==0.0) _startTime = System.currentTimeMillis();
		updateProgress(current, max);
		if(reset) {
			Platform.runLater(new Runnable() {
				@Override
				public void run() {
					_percentComplete.setValue("");
					_timeRunning.setValue("");
					_copyingSpeed.setValue("");
				}
			});
		}
		else {
			Platform.runLater(new Runnable() {
				@Override
				public void run() {
					_percentComplete.setValue(String.format("%2.1f %%", progressProperty().doubleValue()*100.0));
					long now = System.currentTimeMillis();
					double runTime = (now-_startTime)/1000.0;
					int min = (int) (runTime/60);
					int s = (int) runTime - (min*60);
					_timeRunning.setValue(String.format("%d:%02d", min, s));
					double speed = current * _chunkSize / runTime;
					String unit = "B/s";
					if(speed>(4096*1024)) {
						unit = "MB/s";
						speed = speed / 1024 / 1024;
					}
					else if(speed>4096) {
						unit = "kB/s";
						speed = speed / 1024;
					}
					_copyingSpeed.setValue(String.format("%1.1f "+unit, speed));
				}
			});
		}
	}
	
	public StringProperty percentCompleteProperty() {
		return _percentComplete;
	}
	public StringProperty timeRunningProperty() {
		return _timeRunning;
	}
	public StringProperty copyingSpeedProperty() {
		return _copyingSpeed;
	}
}
